@@ -1,10 +1,37 @@
 package Net::SFTP::Foreign::Compat;
 
-our $VERSION = '0.02';
+our $VERSION = '0.03';
+
+use warnings;
+use strict;
+use Carp;
 
 use Net::SFTP::Foreign;
 our @ISA = qw(Net::SFTP::Foreign);
 
+my $suplant;
+
+sub import {
+    for my $arg (@_[1..$#_]) {
+	if ($arg eq ':suplant') {
+	    if (!$suplant) {
+		$suplant = 1;
+
+		@Net::SFTP::ISA = qw(Net::SFTP::Foreign::Compat);
+		@Net::SFTP::Attributes::ISA = qw(Net::SFTP::Foreign::Compat::Attributes);
+		@Net::SFTP::Constant::ISA = qw(Net::SFTP::Foreign::Constant);
+
+		$INC{Net::SFTP} = $INC{Net::SFTP::Foreign::Compat};
+		$INC{Net::SFTP::Attributes} = $INC{Net::SFTP::Foreign::Compat};
+		$INC{Net::SFTP::Constants} = $INC{Net::SFTP::Foreign::Constants};
+	    }
+	}
+	else {
+	    croak "invalid import tag '$arg'"
+	}
+    }
+}
+
 sub new {
     my ($class, $host, %opts) = @_;
 
@@ -18,7 +45,7 @@ sub new {
 
     my $sftp = $class->SUPER::new($host, %opts);
 
-    $sftp->{compat_warn} = $warn;
+    $sftp->{_compat_warn} = $warn;
 
     return $sftp;
 
@@ -26,7 +53,7 @@ sub new {
 
 sub _warn {
     my $sftp = shift;
-    if (my $w = $sftp->{compat_warn}) {
+    if (my $w = $sftp->{_compat_warn}) {
 	$w->(@_);
     }
 }
@@ -76,11 +103,17 @@ sub ls {
     my ($sftp, $path, $cb) = @_;
     if ($cb) {
 	$sftp->SUPER::ls($path,
-			 wanted => sub { $cb->($_[1]); 0});
+			 wanted => sub { _rebless_attrs($_[1]->{a});
+					 $cb->($_[1]);
+					 0 } );
 	return ();
     }
     else {
-	return @{$sftp->SUPER::ls($path) || []}
+	if (my $ls = $sftp->SUPER::ls($path)) {
+	    _rebless_attrs($_->{a}) for @$ls;
+	    return @$ls;
+	}
+	return ()
     }
 }
 
@@ -92,11 +125,13 @@ sub do_realpath { shift->realpath(@_) }
 
 sub do_read {
     my $sftp = shift;
+    my $read = $sftp->sftpread(@_);
+    $sftp->_warn_error;
     if (wantarray) {
-	return ($sftp->read(@_), $sftp->status);
+	return ($read, $sftp->status);
     }
     else {
-	return $sftp->read(@_);
+	return $read
     }
 }
 
@@ -110,7 +145,7 @@ sub _gen_do_and_status {
     }
 }
 
-*do_write = _gen_do_and_status('write');
+*do_write = _gen_do_and_status('sftpwrite');
 *do_close = _gen_do_and_status('close');
 *do_setstat = _gen_do_and_status('setstat');
 *do_fsetstat = _gen_do_and_status('fsetstat');
@@ -119,13 +154,22 @@ sub _gen_do_and_status {
 *do_mkdir = _gen_do_and_status('mkdir');
 *do_rmdir = _gen_do_and_status('rmdir');
 
+sub _rebless_attrs {
+    my $a = shift;
+    if ($a) {
+	bless $a,  ( $suplant
+		     ? "Net::SFTP::Attributes"
+		     : "Net::SFTP::Foreign::Attributes::Compat" );
+    }
+    $a;
+}
 
 sub _gen_do_stat {
     my $method = shift;
     return sub {
 	my $sftp = shift;
 	if (my $a = $sftp->$method(@_)) {
-	    return bless $a, "Net::SFTP::Foreign::Attributes::Compat";
+	    return _rebless_attrs($a);
 	}
 	else {
 	    $sftp->_warn_error;
@@ -139,8 +183,6 @@ sub _gen_do_stat {
 *do_stat = _gen_do_stat('stat');
 
 
-
-
 1;
 
 __END__
@@ -156,6 +198,9 @@ Net::SFTP::Foreign::Compat - Adaptor for Net::SFTP compatibility
     $sftp->get("foo", "bar");
     $sftp->put("bar", "baz");
 
+    use Net::SFTP::Foreign::Compat ':suplant';
+    my $sftp = Net::SFTP->new($host);
+
 =head1 DESCRIPTION
 
 This package is a wrapper around L<Net::SFTP::Foreign> that provides
@@ -165,6 +210,11 @@ Methods on this package are identical to those in L<Net::SFTP> except
 that L<Net::SFTP::Foreign::Attributes::Compat> objects have to be used
 instead of L<Net::SFTP::Attributes>.
 
+If the C<:suplant> tag is used, this module installs also wrappers on
+the C<Net::SFTP> and L<Net::SFTP::Attributes> packages so no other
+parts of the program have to modified in order to move from Net::SFTP
+to Net::SFTP::Foreign.
+
 =head1 COPYRIGHT
 
 Copyright (c) 2006 Salvador FandiE<ntilde>o
